/*
// $Id: //guest/julian_hyde/saffron/src/main/saffron/rel/AggInternalTranslator.java#3 $
// Saffron preprocessor and data engine
// Copyright (C) 2002 Julian Hyde <julian.hyde@mail.com>
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330,
// Boston, MA  02111-1307, USA.
//
// See the COPYING file located in the top-level-directory of
// the archive of this library for complete text of license.
*/

package saffron.rel;

import openjava.mop.*;
import openjava.ptree.*;
import saffron.AggregationExtender;
import saffron.BuiltinAggregation;
import saffron.util.Util;

import java.util.Vector;

/**
 * Converts expressions to consist only of constants, references to group
 * by expressions (variables called "$group0", etc.), and calls to
 * aggregate functions (variables called "$agg0", etc.).
 *
 * <p> These names exist only fleetingly, before {@link AggUnpickler}
 * converts them to field references in the output record. But if we did
 * not use them, we could not be sure whether a field reference such as
 * $input0.$f1 had been converted.
 *
 * <p> Throws {@link NotAGroupException}.
 *
 * @see AggUnpickler
 **/
class AggInternalTranslator extends InternalTranslator
{
    Expression[] groups;
    ExpressionList aggInputList;
    Vector aggCallVector;
    InternalTranslator nonAggTranslator;

    AggInternalTranslator(
        QueryInfo queryInfo, Rel[] inputs, Expression[] groups,
        ExpressionList aggInputList, Vector aggCallVector)
    {
        super(queryInfo, inputs);
        this.groups = groups;
        this.aggInputList = aggInputList;
        this.aggCallVector = aggCallVector;
        this.nonAggTranslator = new InternalTranslator(queryInfo, inputs);
    }

    private Expression toGroupReference(Expression expression, boolean fail)
    {
        for (int i = 0; i < groups.length; i++) {
            if (groups[i].equals(expression)) {
                return new Variable("$group" + i);
            }
        }
		if (fail) {
			throw new NotAGroupException(
				"expression " + expression + " is not a group expression");
		} else {
			return expression;
		}
    }

    // overrides EvaluationShuttle
    public Expression evaluateDown(MethodCall call)
        throws ParseTreeException
    {
        // todo: Unify aggregation lookup with method lookup -- perhaps create
        // a special OJClass with a refined getAcceptableMethod() method. Then
        // make Aggregation and BuiltinAggregation package-protected again. See
        // MethodCall.pickupMethod().
        Environment env = getEnvironment();
        ExpressionList args = call.getArguments();
        int argCount = args.size();
        OJClass[] argTypes = new OJClass[argCount];
        try {
            for (int i = 0; i < argTypes.length; i++) {
                argTypes[i] = args.get(i).getType(env);
            }
        } catch (Exception e) {
            throw Util.newInternal(e);
        }
        Aggregation aggregation = lookupBuiltinAggregation(call, argTypes);
        if (aggregation != null) {
            return makeAggExp(aggregation, call.getArguments());
        }
        aggregation = lookupCustomAggregation(call, argTypes);
        if (aggregation != null) {
            return makeAggExp(aggregation, call.getArguments());
        }
        return toGroupReference(call, false);
    }

    /**
     * Creates and returns an aggregation if <code>call</code> is a call to a
     * builtin, otherwise returns null.
     */
    private Aggregation lookupBuiltinAggregation(
            MethodCall call, OJClass[] argTypes) {
        String methodName = call.getName();
        OJMethod method = BuiltinAggregation.lookup(methodName, argTypes);
        if (method == null) {
            return null;
        }
        return BuiltinAggregation.create(method.getName(), argTypes);
    }

    /**
     * Creates and returns an aggregation if <code>call</code> is a call to a
     * custom aggregation, null otherwise.
     */
    private Aggregation lookupCustomAggregation(
            MethodCall call, OJClass[] argTypes)
    {
        if (!call.getName().equals(AggregationExtender.METHOD_AGGREGATE)) {
            return null;
        }
        OJMethod method;
        try {
            method = call.resolve(getEnvironment());
        } catch (Exception e) {
            return null;
        }
        if (method.getModifiers().isStatic()) {
            return null;
        }
        OJClass reftype = method.getDeclaringClass();
        if (!Toolbox.clazzAggregationExtender.isAssignableFrom(reftype)) {
            return null;
        }
        return new ExtenderAggregation(
			call.getReferenceExpr(), getEnvironment(), argTypes);
    }

    private Expression makeAggExp(Aggregation aggregation, ExpressionList args)
		throws ParseTreeException {
        // translate the arguments into internal form, then canonize them
        args = (ExpressionList) args.makeRecursiveCopy();
        nonAggTranslator.visit(args);
        int argCount = args.size();
        int[] argIndexes = new int[argCount];
outer:
		for (int i = 0; i < argCount; i++) {
			Expression arg = args.get(i);
			for (int j = 0, m = aggInputList.size(); j < m; j++) {
				if (aggInputList.get(j).equals(arg)) {
					// expression already exists; use that
					argIndexes[i] = j;
					continue outer;
				}
			}
			argIndexes[i] = aggInputList.size();
			aggInputList.add(arg);
		}
        Aggregate.Call aggCall = new Aggregate.Call(aggregation, argIndexes);
        // create a new aggregate call, if there isn't already an
        // identical one
        int k = aggCallVector.indexOf(aggCall);
        if (k < 0) {
            k = aggCallVector.size();
            aggCallVector.addElement(aggCall);
        }
        return new Variable("$agg" +  k);
    }

    public Expression evaluateDown(FieldAccess p) {
        return toGroupReference(p,false);
    }
    public Expression evaluateDown(UnaryExpression p) {
        return toGroupReference(p,false);
    }
    public Expression evaluateDown(BinaryExpression p) {
        return toGroupReference(p,false);
    }
    public Expression evaluateDown(ConditionalExpression p) {
        return toGroupReference(p,false);
    }
    public Expression evaluateDown(Variable p) {
        return toGroupReference(p,true);
    }
/*
    public Expression evaluateUp(FieldAccess p) {
        return toGroupReference(p);
    }
    public Expression evaluateUp(UnaryExpression p) {
        return toGroupReference(p);
    }
    public Expression evaluateUp(BinaryExpression p) {
        return toGroupReference(p);
    }
    public Expression evaluateDown(Variable p) {
        Expression e = toGroupReference(p);
        if (e != p) {
            return e;
        }
        throw new NotAGroupException(
            "variable " + p + " is not a group expression");
    }
*/
    boolean isGoodAgg_obsolete(Expression expression)
    {
        if (expression instanceof Literal) {
            return true;
        }
        for (int i = 0; i < groups.length; i++) {
            if (groups[i].equals(expression)) {
                return true;
            }
        }
        if (expression instanceof MethodCall) {
            MethodCall methodCall = (MethodCall) expression;
//				if (isAggregate(methodCall)) {
                return true;
//				}
        }
        if (expression instanceof Variable) {
            Variable variable = (Variable) expression;
            String name = variable.toString();
            QueryEnvironment qenv = (QueryEnvironment)
                this.getEnvironment();
            Expression fromExpression = qenv.lookupFrom(name);
            if (fromExpression != null) {
                // found in current query's from list -- bad
                return false;
            } else {
                // either an alias from an enclosing query, or a java
                // variable -- good
                return true;
            }
        }
        if (expression instanceof NonLeaf) {
            throw Util.newInternal("todo:");
        }
        return false;
    }
}
