/*
// $Id: //guest/julian_hyde/saffron/src/main/saffron/rel/convert/JavaConverter.java#4 $
// Saffron preprocessor and data engine
// Copyright (C) 2002 Julian Hyde <julian.hyde@mail.com>
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330,
// Boston, MA  02111-1307, USA.
//
// See the COPYING file located in the top-level-directory of
// the archive of this library for complete text of license.
*/

package saffron.rel.convert;

import openjava.mop.OJClass;
import openjava.mop.Toolbox;
import openjava.ptree.*;
import saffron.opt.CallingConvention;
import saffron.opt.Cluster;
import saffron.opt.Implementor;
import saffron.rel.Rel;
import saffron.util.Util;

/**
 * <code>JavaConverter</code> converts a plan from <code>inConvention</code> to
 * {@link saffron.opt.CallingConvention#JAVA}.
 **/
class JavaConverter extends Converter
{
    JavaConverter(Cluster cluster, Rel child, int inConvention)
    {
        super(cluster, child, inConvention);
    }

    public Object clone()
    {
        return new JavaConverter(cluster, child, inConvention);
    }

    // implement Rel
    public int getConvention()
    {
        return CallingConvention.JAVA;
    }

    public static boolean canConvert(int inConvention)
    {
        switch (inConvention) {
        case CallingConvention.ARRAY:
        case CallingConvention.ITERATOR:
        case CallingConvention.ENUMERATION:
        case CallingConvention.VECTOR:
        case CallingConvention.MAP:
        case CallingConvention.HASHTABLE:
            return true;
        case CallingConvention.COLLECTION: // do this via Iterator
        default:
            return false;
        }
    }

    // implement Converter
    protected Object implementArray(Implementor implementor, int ordinal)
    {
        switch (ordinal) {
        case -1:
            // Generate
            //   V[] array = <<exp>>;
            //   for (int i = 0; i < array.length; i++) {
            //     V row = array[i];
            //     <<parent>>
            //   }
            StatementList stmtList = implementor.getStatementList();
            Variable variable_array = implementor.newVariable(),
                variable_i = implementor.newVariable();
            StatementList forBody = new StatementList();
            Expression exp = (Expression) implementor.implementChild(
                this, 0, child);
            stmtList.add(
                new VariableDeclaration(
                    TypeName.forOJClass(getType(exp)),
                    variable_array.toString(),
                    exp));
            stmtList.add(
                new ForStatement(
                    new TypeName("int"),
                    new VariableDeclarator[] {
                        new VariableDeclarator(
                            variable_i.toString(),
                            Literal.constantZero())},
                    new BinaryExpression(
                        variable_i,
                        BinaryExpression.LESS,
                        new FieldAccess(
                            variable_array,
                            "length")),
                    new ExpressionList(
                        new UnaryExpression(
                            UnaryExpression.POST_INCREMENT,
                            variable_i)),
                    forBody));
            Variable variable_row = implementor.bind(
					this, forBody,
                    new ArrayAccess(
                        variable_array,
                        variable_i));
            implementor.generateParentBody(this, forBody);
            return null;

        default:
            throw Util.newInternal(
                "implement: ordinal=" + ordinal);
        }
    }

	// implement Converter
    protected Object implementIterator(
        Implementor implementor, int ordinal)
    {
        switch (ordinal) {
        case -1:
            // Generate
            //   Iterator iter = <<exp>>;
            //   while (iter.hasNext()) {
            //     V row = (Type) iter.next();
            //     <<body>>
            //   }
            //
            StatementList
                stmtList = implementor.getStatementList(),
                whileBody = new StatementList();
            Variable variable_iter = implementor.newVariable();
            Expression exp = (Expression) implementor.implementChild(
                this, 0, child);
            stmtList.add(
                new VariableDeclaration(
                    new TypeName("java.util.Iterator"),
                    variable_iter.toString(),
                    exp));
            stmtList.add(
                new WhileStatement(
                    new MethodCall(
                        variable_iter,
                        "hasNext",
                        null),
                    whileBody));
			Variable variable_row = implementor.bind(
					this, whileBody,
                    Util.castObject(
                        new MethodCall(
                            variable_iter,
                            "next",
                            null),
                        Toolbox.clazzObject,
                        child.getRowType()));
            implementor.generateParentBody(this, whileBody);
            return null;

        default:
            throw Util.newInternal(
                "implement: ordinal=" + ordinal);
        }
    }

	// implement Converter
    protected Object implementIterable(
        Implementor implementor, int ordinal)
    {
        switch (ordinal) {
        case -1:
            // Generate
            //   Iterator iter = <<exp>>.iterator();
            //   while (iter.hasNext()) {
            //     V row = (Type) iter.next();
            //     <<body>>
            //   }
            //
            StatementList
                stmtList = implementor.getStatementList(),
                whileBody = new StatementList();
            Variable variable_iter = implementor.newVariable();
            Expression exp = (Expression) implementor.implementChild(
                this, 0, child);
            stmtList.add(
                new VariableDeclaration(
                    new TypeName("java.util.Iterator"),
                    variable_iter.toString(),
                    exp));
            stmtList.add(
                new WhileStatement(
                    new MethodCall(
                        variable_iter,
                        "hasNext",
                        null),
                    whileBody));
			OJClass rowType = child.getRowType();
			Variable variable_row = implementor.bind(
					this, whileBody,
                    Util.castObject(
                        new MethodCall(
                            variable_iter,
                            "next",
                            null),
                        Toolbox.clazzObject,
                        rowType));
            implementor.generateParentBody(this, whileBody);
            return null;

        default:
            throw Util.newInternal(
                "implement: ordinal=" + ordinal);
        }
    }

    // implement Converter
    public Object implementVector(
        Implementor implementor, int ordinal)
    {
        switch (ordinal) {
        case -1:
            // Generate
            //   for (Enumeration e = <<exp>>.elements(); e.hasMoreElements();) {
            //     Row row = (Row) e.nextElement();
            //     <<parent>>
            //   }
            StatementList stmtList = implementor.getStatementList();
            Variable variable_enum = implementor.newVariable();
            StatementList forBody = new StatementList();
            Expression exp = (Expression) implementor.implementChild(
                this, 0, child);
            OJClass rowType = child.getRowType();
            stmtList.add(
                new ForStatement(
                    TypeName.forOJClass(Util.clazzEnumeration),
                    new VariableDeclarator[] {
                        new VariableDeclarator(
                            variable_enum.toString(),
                            new MethodCall(
                                exp,
                                "elements",
                                new ExpressionList()))},
                    new MethodCall(
                        variable_enum,
                        "hasMoreElements",
                        null),
                    new ExpressionList(),
                    forBody));
			Variable variable_row = implementor.bind(
					this, forBody,
					Util.castObject(
                        new MethodCall(
                            variable_enum,
                            "nextElement",
                            null),
                        Toolbox.clazzObject,
                        rowType));
            implementor.generateParentBody(this, forBody);
            return null;

        default:
            throw Util.newInternal(
                "implement: ordinal=" + ordinal);
        }
    }

    // implement Converter
    public Object implementEnumeration(
        Implementor implementor, int ordinal)
    {
        switch (ordinal) {
        case -1:
            // Generate
            //   for (Enumeration enum = <<exp>>; enum.hasMoreElements();)
            //   {
            //     Row row = enum.nextElement();
            //     <<parent>>
            //   }
            StatementList stmtList = implementor.getStatementList();
            Variable variable_enum = implementor.newVariable();
            StatementList forBody = new StatementList();
            Expression exp = (Expression) implementor.implementChild(
                this, 0, child);
            stmtList.add(
                new ForStatement(
                    TypeName.forOJClass(Util.clazzEnumeration),
                    new VariableDeclarator[] {
                        new VariableDeclarator(
                            variable_enum.toString(),
                            exp)},
                    new MethodCall(
                        variable_enum,
                        "hasMoreElements",
                        null),
                    new ExpressionList(),
                    forBody));
			Variable variable_row = implementor.bind(
					this, forBody,
					new MethodCall(
						variable_enum,
						"nextElement",
						null));
            implementor.generateParentBody(this, forBody);
            return null;

        default:
            throw Util.newInternal(
                "implement: ordinal=" + ordinal);
        }
    }

    // implement Converter
    protected Object implementMap(
        Implementor implementor, int ordinal)
    {
        switch (ordinal) {
        case -1:
            // Generate
            //   Map m = <<exp>>;
            //   for (Iterator entries = m.entrySet().iterator();
            //        entries.hasNext();) {
            //     Map.Entry entry = (Map.Entry) keys.next();
            //     Row row = new Row(entry.getKey(), entry.getValue());
            //     <<parent>>
            //   }
            StatementList stmtList = implementor.getStatementList();
            Variable variable_m = implementor.newVariable(),
                variable_entries = implementor.newVariable(),
                variable_entry = implementor.newVariable();
            StatementList forBody = new StatementList();
            Expression exp = (Expression) implementor.implementChild(
                this, 0, child);
            stmtList.add(
                new VariableDeclaration(
                    TypeName.forOJClass(Util.clazzMap),
                    variable_m.toString(),
                    exp));
            stmtList.add(
                new ForStatement(
                    TypeName.forOJClass(Util.clazzIterator),
                    new VariableDeclarator[] {
                        new VariableDeclarator(
                            variable_entries.toString(),
                            new MethodCall(
                                new MethodCall(
                                    variable_m,
                                    "entrySet",
                                    null),
                                "iterator",
                                null))},
                    new MethodCall(
                        variable_entries,
                        "hasNext",
                        null),
                    new ExpressionList(),
                    forBody));
            forBody.add(
                new VariableDeclaration(
                    TypeName.forOJClass(Toolbox.clazzMapEntry),
                    new VariableDeclarator(
                        variable_entry.toString(),
                        new CastExpression(
                            TypeName.forOJClass(Toolbox.clazzMapEntry),
                            new MethodCall(
                                variable_entries,
                                "next",
                                null)))));
			OJClass rowType = getRowType();
			Variable variable_row = implementor.bind(
					this, forBody,
                    new AllocationExpression(
                        TypeName.forOJClass(rowType),
                        new ExpressionList(
                            new MethodCall(
                                variable_entry,
                                "getKey",
                                null),
                            new MethodCall(
                                variable_entry,
                                "getValue",
                                null))));
            implementor.generateParentBody(this, forBody);
            return null;

        default:
            throw Util.newInternal(
                "implement: ordinal=" + ordinal);
        }
    }

    // implement Converter
    protected Object implementHashtable(
        Implementor implementor, int ordinal)
    {
        switch (ordinal) {
        case -1:
            // Generate
            //   Hashtable h = <<exp>>;
            //   for (Enumeration keys = h.keys(); keys.hasMoreElements();) {
            //     Object key = keys.nextElement();
            //     Object value = h.get(key);
            //     Row row = new Row(key, value);
            //     <<parent>>
            //   }
            StatementList stmtList = implementor.getStatementList();
            Variable variable_h = implementor.newVariable(),
                variable_keys = implementor.newVariable(),
                variable_key = implementor.newVariable(),
                variable_value = implementor.newVariable();
            StatementList forBody = new StatementList();
            Expression exp = (Expression) implementor.implementChild(
                this, 0, child);
            stmtList.add(
                new VariableDeclaration(
                    TypeName.forOJClass(Util.clazzHashtable),
                    variable_h.toString(),
                    exp));
            stmtList.add(
                new ForStatement(
                    TypeName.forOJClass(Util.clazzEnumeration),
                    new VariableDeclarator[] {
                        new VariableDeclarator(
                            variable_keys.toString(),
                            new MethodCall(
                                variable_h,
                                "keys",
                                null))},
                    new MethodCall(
                        variable_keys,
                        "hasMoreElements",
                        null),
                    new ExpressionList(),
                    forBody));
            forBody.add(
                new VariableDeclaration(
                    TypeName.forOJClass(Toolbox.clazzObject),
                    new VariableDeclarator(
                        variable_key.toString(),
                        new MethodCall(
                            variable_keys,
                            "nextElement",
                            null))));
            forBody.add(
                new VariableDeclaration(
                    TypeName.forOJClass(Toolbox.clazzObject),
                    new VariableDeclarator(
                        variable_value.toString(),
                        new MethodCall(
                            variable_h,
                            "get",
                            new ExpressionList(variable_key)))));
			OJClass rowType = getRowType();
			Variable variable_row = implementor.bind(
					this, forBody,
                    new AllocationExpression(
                        TypeName.forOJClass(rowType),
                        new ExpressionList(
                            variable_key,
                            variable_value)));
            implementor.generateParentBody(this, forBody);
            return null;

        default:
            throw Util.newInternal(
                "implement: ordinal=" + ordinal);
        }
    }
}
