/*
// $Id: //guest/julian_hyde/saffron/src/main/saffron/rel/plan/PlanNestedLoopJoin.java#2 $
// Saffron preprocessor and data engine
// Copyright (C) 2002 Julian Hyde <julian.hyde@mail.com>
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
// 
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330,
// Boston, MA  02111-1307, USA.
// 
// See the COPYING file located in the top-level-directory of
// the archive of this library for complete text of license.
*/

package saffron.rel.plan;

import openjava.ptree.Expression;
import saffron.Planner;
import saffron.opt.CallingConvention;
import saffron.opt.Cluster;
import saffron.opt.Cost;
import saffron.opt.Implementor;
import saffron.rel.Join;
import saffron.rel.Rel;
import saffron.runtime.Plan;
import saffron.util.Util;

/**
 * <code>PlanNestedLoopJoin</code> implements {@link Join} using the
 * nested-loop algorithm, with output as a {@link saffron.runtime.Plan}.
 *
 * <p>Nested-loops are the simplest implementation of a join: for each row
 * from the left hand side, get each row from the right hand side, apply
 * the condition, and return a joined row if the condition evaluates
 * true.</p>
 *
 * <p>A join isn't a relational expression as such.  The output of a join
 * plan is an array of all of the objects which went into it.  So, for
 * example, the output of <code>emp join (dept join division)</code>
 * contains 3 objects.
 **/
public class PlanNestedLoopJoin extends Join
{
    public PlanNestedLoopJoin(
        Cluster cluster, Rel left, Rel right, Expression condition,
        int joinType)
    {
        super(cluster, left, right, condition, joinType);
    }

    // implement Rel
    public Object clone()
    {
        return new PlanNestedLoopJoin(
            cluster, Util.clone(left), Util.clone(right),
            Util.clone(condition), joinType);
    }

    // implement Plan
    public Cost computeSelfCost(Planner planner)
    {
        double dRows = left.getRows() * right.getRows() *
            Util.getSelectivity(condition),
            dCpu = left.getRows() * right.getRows(),
            dIo = 0;
        return planner.makeCost(dRows, dCpu, dIo);
    }

    // implement Rel
    public int getConvention()
    {
        return CallingConvention.PLAN;
    }

    public Object implement(Implementor implementor, int ordinal)
    {
        Plan leftPlan = (Plan) implementor.implementChild(this, 0, left),
            rightPlan = (Plan) implementor.implementChild(this, 1, right);
        Expression conditionPlan = implementExp(
            condition, new Plan[] {leftPlan, rightPlan});
        return new NestedLoopJoinPlan(
            this, leftPlan, rightPlan, conditionPlan);
    }
}
